#define SDL_MAIN_HANDLED

#include <SDL2/SDL.h>
#include <GL/gl.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

const GLdouble pi = 3.1415926535897932384626433832795;

struct vec2     { float x, y; };
struct vec3     { float x, y, z; };
struct vec4     { float x, y, z, w; };
struct TCamera  { struct vec3 pos, rot; };
struct TMouse   { int x, y, btn; };

class App {

protected:

    SDL_Event       event;
    SDL_Window*     window;
    SDL_GLContext   glcontext;
    int             pticks = 0, width, height;
    int             frame_length = (1000 / 60);
    struct TCamera  camera;
    struct TMouse   mouse;
    int             keyb[256];

public:

    App(int w, int h, const char* title = "OpenGL SDL2 Demo") {

        SDL_Init(SDL_INIT_EVERYTHING);

        window    = SDL_CreateWindow(title, SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, w, h, SDL_WINDOW_OPENGL);
        glcontext = SDL_GL_CreateContext(window);

        // Настройки GL
        glEnable        (GL_DEPTH_TEST);
        glDepthFunc     (GL_LESS);          // Z-буфер, который проверяет тип LESS
        glClearDepth    (1.0f);             // Включить Z-буфер
        glBlendFunc     (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); // Режим смешивания ALPHA
        glBlendEquation (GL_FUNC_ADD);      // Смешивание путем сложения цветов
        glEnable        (GL_TEXTURE_2D);    // Включить текстурирование
        glEnable        (GL_BLEND);         // Включить смешивание
        glEnable        (GL_CULL_FACE);     // Включить срез невидимых областей
        glCullFace      (GL_FRONT);         // Показать передние грани
        glFrontFace     (GL_CCW);           // Тип определения передной грани: против часовой стрелки

        camera.pos = {0, 0, 0};
        camera.rot = {0, 0, 0};
        mouse.btn  = 0;
        width      = w;
        height     = h;

        for (int i = 0; i < 256; i++) keyb[i] = 0;

        updateCamera();
    }

    // Обработка событий
    int poll() {

        while (true) {

            Uint32 ticks = SDL_GetTicks();

            int keypress = 0;

            while (SDL_PollEvent(& event) != 0) {

                switch (event.type) {

                    case SDL_QUIT: {
                        return 0;
                    }

                    // https://wiki.libsdl.org/SDL_MouseMotionEvent
                    case SDL_MOUSEMOTION: {

                        // Движение камеры
                        if (mouse.btn & 1) {
                            camera.rot.y += (event.motion.x - mouse.x) * 0.3;
                            camera.rot.x -= (event.motion.y - mouse.y) * 0.3;
                        }

                        mouse.x = event.motion.x;
                        mouse.y = event.motion.y;
                        break;
                    }

                    // https://wiki.libsdl.org/SDL_MouseButtonEvent
                    case SDL_MOUSEBUTTONDOWN: {

                        mouse.x    = event.motion.x;
                        mouse.y    = event.motion.y;
                        mouse.btn |= event.button.button;
                        break;
                    }

                    case SDL_MOUSEBUTTONUP: {

                        mouse.btn &= ~event.button.button;
                        break;
                    }

                    case SDL_KEYDOWN:

                        keypress = event.key.keysym.scancode;
                        if (keypress < 256) keyb[keypress] = 1;
                        break;

                    case SDL_KEYUP:

                        keypress = event.key.keysym.scancode;
                        if (keypress < 256) keyb[keypress] = 0;
                        break;
                }
            }

            // Обновление экрана
            if (ticks - pticks >= frame_length) {

                pticks = ticks;
                SDL_GL_SwapWindow(window);
                return 1;
            }

            SDL_Delay(1);
        }
    }

    // Выход из SDL
    void destroy() {

        SDL_GL_DeleteContext(glcontext);
        SDL_Quit();
    }

    // Создание и обновление позиции камеры
    void updateCamera() {

        float fovY   = 90,
              zNear  = 0.1;

        GLdouble aspect = width / height;
        GLdouble fH = tan(fovY / 360 * pi) * zNear;
        GLdouble fW = fH * aspect;

        // Загрузка камеры
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        glFrustum(-fW, fW, -fH, fH, zNear, 1000);
        glDepthRange(0.1, 1000);

        // Сначала камера отодвигается, потом вращается по Z,Y,X
        glRotatef(camera.rot.x, -1, 0, 0);
        glRotatef(camera.rot.y,  0, 1, 0);
        glRotatef(camera.rot.z,  0, 0, 1);
        glTranslatef(-camera.pos.x, -camera.pos.y, camera.pos.z);

        // К матрице моделей
        glMatrixMode(GL_MODELVIEW);
    }

    // Установка позиции
    void setCameraPos(vec3 p) { camera.pos = p; }
    void setCameraRot(vec3 r) { camera.rot = r; }

    // Очистка окна от всего что там есть
    void clear(uint cl) {

        int r_ = (cl >> 16) & 255,
            g_ = (cl >>  8) & 255,
            b_ = (cl      ) & 255;

        glClearColor((GLfloat)r_/255., (GLfloat)g_/255., (GLfloat)b_/255., 1);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    }

    // Статус нажатой клавиши
    int pressed(int key) { return keyb[key & 255]; }

    // Движение камеры
    void moveCamera(vec3 v) {

        camera.pos.x += v.x;
        camera.pos.y += v.y;
        camera.pos.z += v.z;
    }

    // Рассчитать вектор, куда смотрит камера
    vec3 lookCamera(float k = 1.) {

        float x = sin(pi * camera.rot.y / 180) * k;
        float y = sin(pi * camera.rot.x / 180) * k;
        float z = cos(pi * camera.rot.y / 180) * k;

        return {x, y, z};
    }

    // Скалярное произведение
    float scalar(vec3 a, vec3 b) {
        return a.x*b.x + a.y*b.y + a.z*b.z;
    }

    // Нормализация вектор
    vec3 normalize(vec3 a, float k = 1.) {

        float r = sqrt(scalar(a, a));
        r = r ? k/r : k;
        return {r*a.x, r*a.y, r*a.z};
    }

    // Текстура создается в памяти, назначается GLuint
    GLuint loadTexture(GLuint* data, int w, int h) {

        GLuint texture_id;

        glGenTextures   (1, (GLuint*) & texture_id);
        glBindTexture   (GL_TEXTURE_2D, texture_id);
        glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        glTexImage2D    (GL_TEXTURE_2D, 0, GL_RGBA, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);

        return texture_id;
    }
};
